/*
* Copyright 2009-2010, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef KYTEA_H__
#define KYTEA_H__

#include "kytea-config.h"
#include "kytea-struct.h"
#include "kytea-model.h"
#include "kytea-lm.h"
#include "dictionary.h"

namespace kytea  {

// a class representing the main analyzer
class Kytea {

private:
    typedef unsigned FeatureId;
    typedef std::list<KyteaSentence*> Sentences;
    typedef std::vector< std::vector< FeatureId > > SentenceFeatures;

    StringUtil* util_;
    KyteaConfig* config_;
    Dictionary * dict_;
    Dictionary * subwordDict_;
    Sentences sentences_;

    KyteaModel* wsModel_;
    KyteaLM * subwordModel_;

    // functions to create dictionaries
    void buildVocabulary();
    
    // a function that checks to make sure that configuration is correct before
    //  training
    void trainSanityCheck();

    // functions for word segmentation
    void trainWS();
    void prepareDictPrefixes(std::vector<unsigned> & dictFeats);
    void prepareNgramPrefixes(std::vector<KyteaString> & charPrefixes, std::vector<KyteaString> & typePrefixes);
    unsigned wsDictionaryFeatures(const KyteaString & sent, SentenceFeatures & feat, const std::vector<unsigned> & featNums);
    unsigned wsNgramFeatures(const KyteaString & sent, SentenceFeatures & feat, const std::vector<KyteaString> & prefixes, int n);

    // functions for pronunciation estimation
    void trainPE();
    unsigned peCharFeatures(const KyteaString & chars, std::vector<unsigned> & feat, const std::vector<KyteaString> & prefixes, KyteaModel * model, int n, int sc, int ec);

    // functions for unknown word PE
    void trainUnk();

    void writeModels();
    // void writeModel(ofstream & out, const KyteaModel & mod);

    void readModels();
    void analyzeInput();
    void calculateWS(KyteaSentence * sent, const std::vector<unsigned> & dictFeats, const std::vector<KyteaString> & charPrefixes, const std::vector<KyteaString> & typePrefixes);
    
    std::vector<KyteaPronunciation> generatePronCandidates(const KyteaString & str);
    void calculateUnknownPE(KyteaWord & str);
    void calculatePE(KyteaSentence * sent, const std::vector<KyteaString> & charPrefixes, const std::vector<KyteaString> & typePrefixes);

public:

    Kytea(KyteaConfig & config) : dict_(0), subwordDict_(0), wsModel_(0), subwordModel_(0) {
        config_ = &config;
        util_ = config.getStringUtil();
        dict_ = new Dictionary(util_);
        subwordDict_ = new Dictionary(util_);
    }
    
    ~Kytea() {
        if(dict_) delete dict_;
        if(subwordDict_) delete subwordDict_;
        if(wsModel_) delete wsModel_;
        if(subwordModel_) delete subwordModel_;
        for(Sentences::iterator it = sentences_.begin(); it != sentences_.end(); it++)
            delete *it;
    }

    // train the analyzer
    void trainAll() {
        
        // sanity check
        trainSanityCheck();
        
        // load the vocabulary, pronunciations
        buildVocabulary();

        // train the word segmenter
        if(config_->getDoWS())
            trainWS();
        
        // train the pronunciation estimator
        if(config_->getDoPE()) {
            trainPE();
            if(config_->getDoUnk() && config_->getSubwordDictFiles().size() > 0)
                trainUnk();
        }

        // write the models out to a file
        writeModels();

    }

    void analyze() {
        
        // on full input, disable word segmentation
        if(config_->getInputFormat() == CorpusIO::FORMAT_FULL)
            config_->setDoWS(false);

        // sanity check
        std::ostringstream buff;
        if(config_->getModelFile().length() == 0)
            throw std::runtime_error("A model file must be specified to run Kytea (-model)");
        
        // read the models in from the model file
        readModels();
        if(!config_->getDoWS() && !config_->getDoPE()) {
            buff << "Both word segmentation and pronunciation estimation are disabled." << std::endl
                 << "At least one must be selected to perform processing." << std::endl;
            throw std::runtime_error(buff.str());
        }
        // set the input format
        if(config_->getDoWS()) {
            if(config_->getInputFormat() == CorpusIO::FORMAT_DEFAULT)
               config_->setInputFormat(CorpusIO::FORMAT_RAW);
        } else {
            if(config_->getInputFormat() == CorpusIO::FORMAT_DEFAULT)
                 config_->setInputFormat(CorpusIO::FORMAT_FULL);
            else if(config_->getInputFormat() == CorpusIO::FORMAT_RAW) {
                buff << "In order to handle raw corpus input, word segmentation must be turned on." << std::endl
                     << "Either specify -in {full,part,prob}, stop using -nows, or train a new " << std::endl
                     << "model that has word segmentation included." << std::endl;
                throw std::runtime_error(buff.str());
            }
        }

        // parse the sentences one by one
        analyzeInput();

    }

};

}

#endif
